#ifndef _P1_CAMERA_LIVEVIEW_CONFIGURATION_
#define _P1_CAMERA_LIVEVIEW_CONFIGURATION_

#include <memory>
#include <string>

#include "C_P1CameraCommonTypes.h"
#include "C_P1CameraCommonLinkage.h"

//Forward decl
class LiveViewConfigurationPimpl;

namespace P1
{
namespace CameraSdk
{
    /**
     * @brief Representation of a cameras curretn LiveView config (setup)
     *
     * This class allows you to configure the LiveView properties of a camera.
     *
     * You obtain an object of this class using the @ref Camera::GetLiveViewConfiguration on the Camera.
     * As you edit the parameters on the class, the corrospanding values are changed in the camera.
     *
     * **Note**: Changing the LiveView configuration is an exspensive operation, and requires a lot of work
     * to be done in the camera. Also, the LiveView image feed is heavily pipelined, so changes will occur
     * after a some delay.
     *
     * The @ref LiveViewImage class helps you account of this, by including the current LiveView image's
     * *crop* parameters.
     *
     * This LiveVIew output image size is independent from the applied crop. It is only influenced by the
     * @ref PreferredMaxDimension setting. This defines the length of the biggest of the image dimensions,
     * which is most likely the width.
     *
     * The default LiveView image width is 1024 pixels, but you can change it to other common resolutions;
     * like 640, 800, etc. You cannot set it to an arbitrary integer. The desired dimension value will snap the
     * supported image resolutions.
     */
    class LiveViewConfiguration
    {
    private:
        std::unique_ptr<LiveViewConfigurationPimpl> mImpl;

    public:

        EXPORT_API LiveViewConfiguration(LiveViewConfigurationPimpl* pimpl);

        // Getters

        EXPORT_API LiveViewConfiguration(const LiveViewConfiguration& other) = delete;
        EXPORT_API LiveViewConfiguration& operator=(const LiveViewConfiguration& other) = delete;

        // Move constructor allowed, since an owning Camera can be moved
        EXPORT_API LiveViewConfiguration(LiveViewConfiguration&& other);

        /** @brief Get the current bitmap format setting of LiveView */
        EXPORT_API LiveViewBitmapFormat Format() const;
        
        /** @brief Get the current crop applied to the LiveView */
        EXPORT_API ValueRect Crop() const;
        
        /** @brief Get the current desired max output dimension of the LiveView */
        EXPORT_API uint32_t PreferredMaxDimension() const;

        /** @brief Get the Helper method to reset the crop to "no crop" */
        EXPORT_API ValueArea FullImageSize() const;

        /**
         * @brief An abstract value for exposure compensation
         *
         * This values maps to the exposure compensation, but does not use the same EV values.
         * Instead it uses an opaque float point value.
         *
         * @return float The current lightness
         */
        EXPORT_API float Lightness() const;

        // Setters

        /** @brief Set the LiveView image's bitmap format */
        EXPORT_API void SetFormat(const LiveViewBitmapFormat format);
        
        /** @brief Set a new crop on the LiveView */
        EXPORT_API void SetCrop(const ValueRect& cropRect);
        
        /** @brief Set the desired maximum dimension (width) of the LiveView's output images */
        EXPORT_API void SetPreferredMaxDimension(const uint32_t maxDimensionValue);
        
        /**
         * @brief Set the *Custom C1* white balance, by the sampling a  position in full image coordinates
         *
         * This method changes the cameras white balance property to *Custom C1* setting. And the
         * white balance vector for *Custom C1* is changed to reflect the sample position provided by this
         * method. **This means the cameras white balance setting is changed!**
         *
         * Use the *white balance* property, to change white balance back to *Auto* or *Daylight*, etc.
         */
        EXPORT_API void SetWhiteBalancePoint(const ValuePoint& whitePoint);
        
        /** @brief Set a *exposure compensation*-like  value applied to the live view */
        EXPORT_API void SetLightness(const float lightness);

        EXPORT_API std::string ToString() const;
    };

} // CameraSDK
} // P1

#endif
