#ifndef P1CAMERA_TYPES_HPP
#define P1CAMERA_TYPES_HPP

#include <cstdint>
#include <string>
#include <vector>
#include <functional>

#include "C_P1CameraCommonTypes.h"
#include "C_P1CameraCommonLinkage.h"

namespace P1
{
namespace CameraSdk
{

/**
 * @brief Hold a single property value, and acts like a wrapper of different property types
 *
 * This class holds only the value of a property. It does not know about the property itself,
 * for that we have the @ref PropertySpecification class. PropertyValue can hold value of type:
 * `integer`, `floating point`, `string`, `point`, `color`, `area` and `rect`.
 *
 * A propertyValue also have a @ref PropertyValue::mPresentationString, that is a string
 * representation of its value.
 *
 * Preferrably you should not create PropertyValues yourself. Rather you should obtains them
 * from a camera and modify them.
 *
 * If you are familiar with *Qt*, the `QVariant` class achieves the same goal is this class.
 *
 * @sa PropertyType
 * @sa Camera::GetProperty
 */
struct EXPORT_API PropertyValue
{
    /** @brief Create an empty property value (no data) */
	PropertyValue() { }

    /** @brief Create a string type property value, based on the provided string */
	PropertyValue(const std::string& fromString);

    /** @brief Create a Int64 type property value, based on the provided integer */
	PropertyValue(const int64_t& fromInteger);

    /** @brief Create a Int32 type property value, based on the provided integer */
	PropertyValue(const int32_t& fromInteger);

    /** @brief Create a float64 type property value, based on the provided double */
	PropertyValue(const double& fromDouble);

    /** @brief Indicate the type of the value contained in this PropertyValue object */
	PropertyType mType = kPropertyTypeUndefined;

    /** @brief All integer based values are contained in this member (enum, int, uint, time, etc.) */
    int64_t mInt = 0;

    /** @brief All floating point values are contained in this member. (float 32 and 64) */
	double mDouble = 0.0;

    /** @brief If this PropertyValue is a string type, the string is contained here */
	std::string mString;

    /** @brief If this PropertyValue is a point type, it is contained here */
	ValuePoint mPoint;

    /** @brief If this PropertyValue is a color type, it is contained here */
	ValueColorFloat mColor;

    /** @brief If this PropertyValue is a area type, it is contained here */
	ValueArea mArea;

    /** @brief If this PropertyValue is a floating point area type, it is contained here */
	ValueAreaFloat mAreaFloat;

    /** @brief If this PropertyValue is a rectangle type, it is contained here */
	ValueRect mRect;

    /** @brief If this PropertyValue is a integer range type, it is contained here */
    ValueRange mRange { 0, 0 };

    /** @brief If this PropertyValue is a floating point range type, it is contained here */
    ValueRangeFloat mRangeFloat { 0.0, 0.0 };

    /** @brief This is a human readable representation of the value */
	std::string mPresentationString;

	/**
	 * @brief Returns a simple string representation of this property
	 * @return A string value that describes the PropertyValue's value
	 */
	std::string ToString() const;
	std::string StringValue() const;

    /**
     * @brief Create a enum type property value, based on the provided integer
     */
	static PropertyValue CreateEnum(uint32_t enumValue);
};



// C++ Version of PropertyRange

/**
 * @brief Numeric property _range span_ class, used by PropertySpecification
 *
 * Some camera properties have a valid range of values, defined by a minimum and
 * maximum. This class defines these boundaries.
 *
 * A valid PropertyRange object, _must_ have its member @ref mHas set to `true`.
 * `false` indicates an invalid range object.
 */
struct EXPORT_API PropertyRange
{
    /** @brief Indicate if this object is valid, and range check should be applied */
	bool mHas = false;

    /** @brief The lower bound of the range */
	PropertyValue mMin;

    /** @brief The upper bound of the range */
	PropertyValue mMax;

    /** @brief Get a human readable representation of the range */
	std::string ToString() const;
};

// C++ Version of PropertySpecification

/**
 * @brief A detailed description of a camera property
 *
 * This class defines if a property has any restrictions to its valid values,
 * like a range or even a list of allowed values.
 *
 * All Properties belong to a category or group, have a unit and a visibility
 * level, called the _user level_. The _user level_ defines if the property
 * is an advanced camera setting, that should probably be hidden from novice
 * users.
 *
 * To represent the current value of a proerty we use the class
 * @ref PropertyValue. It is also there the _type_ of the property is defined.
 *
 * You cannot change property specifications, you can change only property
 * values, and only using @ref Camera::SetProperty
 *
 * You can obtain PropertySpecification objects from @ref Camera::GetPropertySpec
 */
struct EXPORT_API PropertySpecification
{
    /**
     * @brief The assigned id of the property
     *
     * All properties are identified uniquely by their _Id_. You use the _Id_ to
     * reference and look up properties on a camera.
     */
	uint32_t mId = kPropertyUndefined;

    /**
     * @brief The current value (and type) of the property
     *
     * This is the current assigned value to the property. If the property if
     * not marked as _read-only_ (@ref mReadOnly), this value can be changed using
     * @ref Camera::SetProperty.
     */
	PropertyValue mValue;

    /**
     * @brief The (human readable) name of the property
     *
     * This name is localized based on the camera's UI language setting. However,
     * the default language is english.
     *
     * This name can be presented to users in a UI
     */
	std::string mName;

    /**
     * @brief The group that this property belongs to
     *
     * The group is a string that might contain double colons "::", to denote
     * subgroups. You can use the _group_ to group related properties in a UI.
     */
	std::string mGroup;

    /**
     * @brief Indicates if this property is currently _read only_
     *
     * Some properties are always _read only_, but others might be _read only_
     * as a result of the camera settings. For example, if the exposure program
     * is set to _auto_, then the _ISO_, _Aperture_ and _shutter_ properties are
     * _read_only_. If the exposure program is set to _manual_, then they are
     * writeable.
     */
	bool mReadOnly = false;

    /**
     * @brief
     *
     *
     */
	bool mEnabled = true;

    /**
     * @brief The visibility level of this peroperty
     *
     * Some properties are more advanced than others. Use this member to hide
     * the advanced properties from novice users.
     */
	PropertyUserLevel mUserLevel = kPropertyUserLevelUndefined;

    /**
     * @brief
     *
     *
     */
	bool mSelectOnly = false;

    /**
     * @brief
     *
     *
     */
	bool mDisplayAsHex = false;

    /**
     * @brief
     *
     *
     */
	bool mRefreshRequired = false;

    /**
     * @brief
     *
     *
     */
	std::string mUnitString;

    /**
     * @brief
     *
     *
     */
	std::string mDescription;

    /**
     * @brief The sorting order of the property
     *
     * Instead of presenting properties as sorted by _Id_ or name, sort by
     * this integer. (Per group.)
     */
	uint32_t mDisplayIndex = 0;

    /**
     * @brief The default value of the property
     *
     * This is the value that the property is restored to, then "reset to
     * default" is triggered, on the camera.
     *
     * _Note_: that not all properties have a meaningful default value.
     * (Such as "System clock")
     */
	PropertyValue mDefault;

    /**
     * @brief The valid range of the property value
     *
     * Some properties are restricted to have a valid range of values they can
     * take.
     * If a range is applicable for this property, the @ref PropertyRange::mHas
     * is `true`. If `false`, this member can be ignored.
     */
	PropertyRange mRange;

    /**
     * @brief List of allowed values, for this property
     *
     * Some properties can only take values, from this list of allowed values.
     * If this list is present (size > 0), and you set a value that is not in
     * the list, two things can happen: 1. the camera rejects the value or, 2.
     * the value is truncated to the nearest value in the list.
     */
	std::vector<PropertyValue> mValueList;

    /**
     * @brief Get a human readable description of this PropertySpecification
     */
	std::string ToString() const;
};

/**
 * @brief Brief description of a camera - without necessarily having to open it
 *
 * This class is used by the @ref Camera::GetAvailableCameras method, to
 * return a list of available cameras. You do not have to create instances of this
 * class yourself.
 *
 * @sa Camera::GetAvailableCameras
 */
struct EXPORT_API CameraListElement
{
	/** The connection type where camera is present */
	TetherConnection mConnection = kTetherConnection_Undefined;

	/** Camera display name */
	std::string mName;

	/** The cameras serial number */
	std::string mSerialNum;

	/** Camera's IP address, if it is available via TCP/IP */
	std::string mIPv4;

	/** Bool indicating if the camera is open by the SDK.
	    Note, this is a snapshot of the state when this camera-list-element
		is created, hence may have changed since. */
	bool mIsOpen = false;
};

class Camera; // forward decl

/**
 * @brief A transaction that can set a property and wait for a change notification
 *
 * Since setting a property is asynchronous by nature, you can use this helper
 * class to chain setting one property after a anothers _change notification_ has
 * arrived.
 *
 * This is particularly useful when setting one property, puts another out of
 * _read only_ state. Then you _must_ wait for the first property to be in effect,
 * before you are allowed to modify the second.
 *
 * This class is purely a helper class, meaning that it uses only the existing
 * public API. It creates a @ref Listener, and then wait for one or more
 * property _change notifications_, before it completes the transaction.
 *
 * In this example we use a transaction to set the _exposure program_ and wait
 * for it to be in effect, before we can set the ISO value:
 *
 * @code{.cpp}
 *  uint32_t exposureProgramId = 1000;
 *  uint32_t isoId = 120;
 *  PropertyTransaction expoProg(camera, exposureProgramId, PropertyValue::CreateEnum(0));
 *  expoProg.AndWaitFor(exposureProgramId); // wait until property is in effect
 *  camera.SetProperty(isoId, PropertyValue(800)); // now, set the ISO value
 * @endcode
 *
 * The @ref Camera class has a method to create these PropertyTransaction objects
 * @ref Camera::BeginSetProperty
 */
class EXPORT_API PropertyTransaction
{
private:
	Camera& mCamera;
	const int mPropertyId;
	const PropertyValue& mValue;

public:
    /**
     * @brief Create a transaction that starts with the provided Property _Id_
     * and value.
     *
     * This constructs the transaction object, but actual setting the property
     * on the camera, does not happen here. It happens when you call one of the
     * multiple `AndWaitFor` methods.
     *
     * @param camera The camera that the property should be applied to
     * @param propertyId The _Id_ of the first property to set
     * @param value The value that the first property is set to
     *
     * @sa AndWaitFor(const uint32_t, const uint32_t)
     */
	PropertyTransaction(Camera& camera, const uint32_t propertyId, const PropertyValue& value);

    /**
     * @brief Begins the transaction by applying the property provided in the
     * constructor, and then wait for a _change notification_ with the provided _Id_
     *
     * First the property provided in the transactions constructor is applied.
     * Then wait for the first property value _change notification_ to arrive
     * for the provided property _Id_.
     *
     * @param propertyId The _Id_ that we wait for a _change notification_ for
     * @param timeout The maximum time we will wait (in MS). 0 is forever
     */
	void AndWaitFor(const uint32_t propertyId, const uint32_t timeout = 1000);

    /**
     * @brief Begins the transaction and wait for a _change notification_, for
     * a property, who's specification fulfills a condition.
     *
     * Applies the property supplied in the constructor and then wait for a value
     * _change notification_. Then validate the PropertySpecification against
     * a callback function.
     *
     * The callback function takes the PropertySpecification that triggered the
     * _change notification_ and must return `true`or `false`. If `true` the call
     * returns. Otherwise, we keep waiting on _change notifications_.
     *
     * You can choose to also validate the PropertySpecification against the
     * callback, _before_ we wait for the _change notification_. This allows you
     * to handle situations, where a _change notification_ never arrives, because
     * you are not changing the property's value. (It already has the value,
     * you are trying to set it to.)
     *
     * @param propertyId The _Id_ that we wait for a _change notification_ for
     * @param condition The callback function, that validate the arriving property changes
     * @param isPrecondition Defaults to `true`, if you want to also validate, _before_ you wait for changes
     * @param timeout An optional timeout in MS, before you give up waiting. 0 is forever
     */
	void AndWaitFor(const uint32_t propertyId, std::function<bool(const PropertySpecification&)> condition, bool isPrecondition = true, const uint32_t timeout = 1000);

    /**
     * @brief Begins the transaction and wait for _change notifications_, for **all**
     * provided property Ids, who's specification fulfills a condition.
     *
     * Applies the property supplied in the constructor and then wait for a value
     * _change notification_ for **all** the provided property _Ids_. Then
     * validate each PropertySpecification against a callback function.
     *
     * The callback function takes the PropertySpecification that triggered the
     * _change notification_ and must return `true`or `false`. When the _condition_
     * callback has returned `true` every provided property _id_, the call returns.
     *
     * You can choose to also validate the PropertySpecifications against the
     * callback, _before_ you wait for the _change notifications_. This allows you
     * to handle situations, where a _change notification_ never arrives, because
     * you are not changing the property's value. (It already has the value,
     * you are trying to set it to.)
     *
     * @param propertyIds The _Ids_ that we wait for _change notifications_ for
     * @param condition The callback function, that validate the arriving property changes
     * @param isPrecondition Defaults to `true`, if you want to also validate, _before_ you wait for changes
     * @param timeout An optional timeout in MS, before you give up waiting. 0 is forever
     */
	void AndWaitForAll(const std::vector<uint32_t>& ids, std::function<bool(const PropertySpecification&)> condition, bool isPrecondition = true, const uint32_t timeout = 1000);
};

} // namespace CameraSdk
} // namespace P1

#endif // P1CAMERA_TYPES_HPP
